# coding=utf-8
from .datastore import DataStore
from .datastore import open_datastore
from .datastore import read_xml_dir

__version__ = '0.1.2'
__all__ = ["DataStore", "open_datastore", "read_xml_dir"]



# coding=utf-8
import os
from pathlib import Path
from xml.etree import ElementTree

import numpy as np
import pandas as pd
import xmltodict


def read_data_from_fp_numpy(fp):
    """
    Read the data from a single Silixa xml file. Using a simple approach

    :param fp: File path
    :param i_first: Index of the first <data>. Use this if reading multiple files.
    :param i_last:  Index of the last </data>. Use this if reading multiple files.
    :return: The data of the file as numpy array of shape (nx, ncols)

    :note: calculating i_first and i_last is fast compared to the rest
    """
    with open(fp) as fh:
        s = fh.readlines()

    s = [si.strip() for si in s]  # remove xml hierarchy spacing

    i_first = s.index('<data>')
    i_last = len(s) - s[::-1].index('</data>') - 1

    lssl = slice(i_first + 1, i_last, 3)  # list of strings slice

    data = np.loadtxt(s[lssl], delimiter=',', dtype=float)

    return data


def metakey(meta, dic_to_parse, prefix, sep):
    for key in dic_to_parse:
        if prefix == "":
            prefix_parse = key.replace('@', '')
        else:
            prefix_parse = sep.join([prefix, key.replace('@', '')])

        if prefix_parse == sep.join(('logData', 'data')):  # u'logData:data':
            continue

        if hasattr(dic_to_parse[key], 'keys'):
            meta.update(metakey(meta, dic_to_parse[key], prefix_parse, sep))

        elif isinstance(dic_to_parse[key], list):
            for ival, val in enumerate(dic_to_parse[key]):
                num_key = prefix_parse + '_' + str(ival)
                meta.update(metakey(meta, val, num_key, sep))
        else:

            meta[prefix_parse] = dic_to_parse[key]

    return meta


# In[]: Define functions
def read_meta(filename, sep):
    with open(filename) as fh:
        doc_ = xmltodict.parse(fh.read())

    if u'wellLogs' in doc_.keys():
        doc = doc_[u'wellLogs'][u'wellLog']
    else:
        doc = doc_[u'logs'][u'log']

    return metakey(dict(), doc, '', sep)


def clear_output_dir(fp, fn='*.nc'):
    for p in Path(fp).glob(fn):
        p.unlink()


def find_dimensions(filename):
    # Returns nitem and nx
    with open(filename) as fh:
        doc_ = xmltodict.parse(fh.read())

    if u'wellLogs' in doc_.keys():
        doc = doc_[u'wellLogs'][u'wellLog']
    else:
        doc = doc_[u'logs'][u'log']

    nitem = len(doc[u'logCurveInfo'])
    nx = len(doc[u'logData']['data'])

    return nitem, nx


def grab_data(filelist, sep=':'):
    ns = {
        's': 'http://www.witsml.org/schemas/1series'}

    # Obtain meta data from the first file
    meta = read_meta(filelist[0], sep)

    nitem, nx = find_dimensions(filelist[0])
    ntime = len(filelist)

    double_ended_flag = bool(int(meta['customData:isDoubleEnded']))
    chFW = int(meta['customData:forwardMeasurementChannel']) - 1  # zero-based

    extra2 = [('log', 'customData', 'acquisitionTime'),
              ('log', 'customData', 'referenceTemperature'),
              ('log', 'customData', 'probe1Temperature'),
              ('log', 'customData', 'probe2Temperature'),
              ('log', 'customData', 'referenceProbeVoltage'),
              ('log', 'customData', 'probe1Voltage'),
              ('log', 'customData', 'probe2Voltage')
              ]

    if not double_ended_flag:
        extra2.append(('log', 'customData', 'UserConfiguration',
                       'ChannelConfiguration', 'AcquisitionConfiguration',
                       'AcquisitionTime', 'userAcquisitionTime'))

    else:
        extra2.append(('log', 'customData', 'UserConfiguration',
                       'ChannelConfiguration', 'AcquisitionConfiguration',
                       'AcquisitionTime', 'userAcquisitionTimeFW'))
        extra2.append(('log', 'customData', 'UserConfiguration',
                       'ChannelConfiguration', 'AcquisitionConfiguration',
                       'AcquisitionTime', 'userAcquisitionTimeBW'))

    extra = {
        item[-1]: dict(path=item, array=np.zeros(ntime, dtype=np.float32))
        for item in extra2
        }
    arr_path = 's:' + '/s:'.join(['log', 'logData', 'data'])
    '''allocate and add units'''
    for key, item in extra.items():
        item['array'] = np.zeros(ntime, dtype=np.float32)

        if f'customData:{key}:uom' in meta:
            item['uom'] = meta[f'customData:{key}:uom']

        else:
            item['uom'] = ''

    # Lookup recorded item names
    ddict = []
    for iitem in range(nitem):
        key = u'logCurveInfo_' + str(iitem) + sep + 'mnemonic'
        ddict.append((str(meta[key]), np.float32))

    # Allocate data and time arrays
    ddtype = np.dtype(ddict)
    array = np.zeros((nx, ntime), dtype=ddtype)
    timearr = np.zeros(
        ntime,
        dtype=[('minDateTimeIndex', '<U29'), ('maxDateTimeIndex', '<U29'),
               ('filename_tstamp', (np.unicode_, 17))])  # 'S17'

    # print summary
    print('%s files were found, each representing a single timestep' % ntime)
    print('%s recorded vars were found: ' % nitem + ', '.join(ddtype.names))
    print('Recorded at %s points along the cable' % nx)

    # grab data from all *.xml files in filelist
    for it, file_ in enumerate(filelist):
        if it % 10 == 0:
            print('processing file ' + str(it + 1) + ' out of ' + str(ntime))

        with open(file_, 'r') as fh:
            tree = ElementTree.parse(fh)

            arr_el = tree.findall(arr_path, namespaces=ns)
            arr_str = [
                tuple(arr_eli.text.strip().split(',')) for arr_eli in arr_el
                ]
            assert len(arr_str) == nx, 'This file has a different nr of measurement points: ' + \
                                       file_
            array[:, it] = arr_str

            for key, item in extra.items():
                if 'ChannelConfiguration' in item['path']:
                    path1 = 's:' + '/s:'.join(item['path'][:4])
                    val1 = tree.findall(path1, namespaces=ns)
                    path2 = 's:' + '/s:'.join(item['path'][4:6])
                    val2 = val1[chFW].find(path2, namespaces=ns)
                    item['array'][it] = val2.text

                else:
                    path = 's:' + '/s:'.join(item['path'])
                    val = tree.find(path, namespaces=ns)
                    item['array'][it] = val.text

            timearr[it]['minDateTimeIndex'] = tree.find(
                's:log/s:startDateTimeIndex', namespaces=ns).text
            timearr[it]['maxDateTimeIndex'] = tree.find(
                's:log/s:endDateTimeIndex', namespaces=ns).text

            file_name = os.path.split(file_)[1]
            timearr[it]['filename_tstamp'] = file_name[10:-4]  # hack

    return array, timearr, meta, extra


def coords_time(double_ended_flag, extra, timearr, timezone_netcdf, timezone_ultima_xml):
    if not double_ended_flag:
        # single ended measurement
        dt1 = pd.to_timedelta(extra['userAcquisitionTimeFW']['array'],
                              's').values

        # start of the forward measurement
        index_time_FWstart = pd.to_datetime(
            (pd.DatetimeIndex(timearr['maxDateTimeIndex']) - dt1).tz_localize(
                tz=timezone_ultima_xml).tz_convert(timezone_netcdf))

        # end of the forward measurement
        index_time_FWend = pd.to_datetime(
            pd.DatetimeIndex(timearr['maxDateTimeIndex']).tz_localize(
                tz=timezone_ultima_xml).tz_convert(timezone_netcdf))

        # center of forward measurement
        index_time_FWmean = pd.to_datetime(
            (pd.DatetimeIndex(timearr['maxDateTimeIndex']) - dt1 / 2).tz_localize(
                tz=timezone_ultima_xml).tz_convert(timezone_netcdf))

        coords = {
            'timestart': index_time_FWstart.astype('datetime64[ns]'),
            'timeend':   index_time_FWend.astype('datetime64[ns]'),
            'time':      index_time_FWmean.astype('datetime64[ns]')}  # in UTC

    else:
        # double ended measurement
        dt1 = pd.to_timedelta(extra['userAcquisitionTimeFW']['array'],
                              's').values
        dt2 = pd.to_timedelta(extra['userAcquisitionTimeBW']['array'],
                              's').values

        # start of the forward measurement
        index_time_FWstart = pd.to_datetime(
            (pd.DatetimeIndex(timearr['maxDateTimeIndex']) - dt1).tz_localize(
                tz=timezone_ultima_xml).tz_convert(timezone_netcdf))

        # end of the forward measurement
        index_time_FWend = pd.to_datetime(
            pd.DatetimeIndex(timearr['maxDateTimeIndex']).tz_localize(
                tz=timezone_ultima_xml).tz_convert(timezone_netcdf))

        # center of forward measurement
        index_time_FWmean = pd.to_datetime(
            (pd.DatetimeIndex(timearr['maxDateTimeIndex']) - dt1 / 2).tz_localize(
                tz=timezone_ultima_xml).tz_convert(timezone_netcdf))

        # start of the backward measurement
        index_time_BWstart = index_time_FWend.copy()

        # end of the backward measurement
        index_time_BWend = pd.to_datetime(
            (pd.DatetimeIndex(timearr['maxDateTimeIndex']) + dt2).tz_localize(
                tz=timezone_ultima_xml).tz_convert(timezone_netcdf))

        # center of backward measurement
        index_time_BWmean = pd.to_datetime(
            (pd.DatetimeIndex(timearr['maxDateTimeIndex']) +
             dt2 / 2).tz_localize(
                tz=timezone_ultima_xml).tz_convert(timezone_netcdf))

        coords = {
            'timeFWstart': index_time_FWstart.astype('datetime64[ns]'),
            'timeFWend':   index_time_FWend.astype('datetime64[ns]'),
            'timeFW':      index_time_FWmean.astype('datetime64[ns]'),
            'timeBWstart': index_time_BWstart.astype('datetime64[ns]'),
            'timeBWend':   index_time_BWend.astype('datetime64[ns]'),
            'timeBW':      index_time_BWmean.astype('datetime64[ns]'),
            'time':        index_time_FWend.astype('datetime64[ns]')}  # in UTC

    return coords



# coding=utf-8
import glob
import inspect
import os

import xarray as xr
import yaml

from dtscalibration.datastore_utils import coords_time
from dtscalibration.datastore_utils import grab_data


class DataStore(xr.Dataset):
    def __init__(self, *args, **kwargs):
        """To load data from a file or file-like object, use the `open_datastore`
        function.

        Parameters
        ----------
        data_vars : dict-like, optional
            A mapping from variable names to :py:class:`~xarray.DataArray`
            objects, :py:class:`~xarray.Variable` objects or tuples of the
            form ``(dims, data[, attrs])`` which can be used as arguments to
            create a new ``Variable``. Each dimension must have the same length
            in all variables in which it appears.
        coords : dict-like, optional
            Another mapping in the same form as the `variables` argument,
            except the each item is saved on the dataset as a "coordinate".
            These variables have an associated meaning: they describe
            constant/fixed/independent quantities, unlike the
            varying/measured/dependent quantities that belong in `variables`.
            Coordinates values may be given by 1-dimensional arrays or scalars,
            in which case `dims` do not need to be supplied: 1D arrays will be
            assumed to give index values along the dimension with the same
            name.
        attrs : dict-like, optional
            Global attributes to save on this dataset.
        sections : dict, optional
            Sections for calibration. The dictionary should contain key-var couples
            in which the key is the name of the calibration temp time series. And
            the var is a list of slice objects as 'slice(start, stop)'; start and
            stop in meter (float).
        compat : {'broadcast_equals', 'equals', 'identical'}, optional
            String indicating how to compare variables of the same name for
            potential conflicts when initializing this dataset:

            - 'broadcast_equals': all values must be equal when variables are
              broadcast against each other to ensure common dimensions.
            - 'equals': all values and dimensions must be the same.
            - 'identical': all values, dimensions and attributes must be the
              same.
        """
        super(DataStore, self).__init__(*args, **kwargs)

        if '_sections' not in self.attrs:
            self.attrs['_sections'] = yaml.dump(None)

        if 'sections' in kwargs:
            self.sections = kwargs['sections']

    @property
    def sections(self):
        assert hasattr(self, '_sections')
        return yaml.load(self.attrs['_sections'])

    @sections.setter
    def sections(self, sections):
        # assert sections
        self.attrs['_sections'] = yaml.dump(sections)
        pass

    @sections.deleter
    def sections(self):
        self.sections = None
        pass


def open_datastore(filename_or_obj, **kwargs):
    """Load and decode a dataset from a file or file-like object
    into a DataStore object.

    Parameters
    ----------
    filename_or_obj : str, Path, file or xarray.backends.*DataStore
        Strings and Path objects are interpreted as a path to a netCDF file
        or an OpenDAP URL and opened with python-netCDF4, unless the filename
        ends with .gz, in which case the file is gunzipped and opened with
        scipy.io.netcdf (only netCDF3 supported). File-like objects are opened
        with scipy.io.netcdf (only netCDF3 supported).
    group : str, optional
        Path to the netCDF4 group in the given file to open (only works for
        netCDF4 files).
    decode_cf : bool, optional
        Whether to decode these variables, assuming they were saved according
        to CF conventions.
    mask_and_scale : bool, optional
        If True, replace array values equal to `_FillValue` with NA and scale
        values according to the formula `original_values * scale_factor +
        add_offset`, where `_FillValue`, `scale_factor` and `add_offset` are
        taken from variable attributes (if they exist).  If the `_FillValue` or
        `missing_value` attribute contains multiple values a warning will be
        issued and all array values matching one of the multiple values will
        be replaced by NA. mask_and_scale defaults to True except for the
        pseudonetcdf backend.
    decode_times : bool, optional
        If True, decode times encoded in the standard NetCDF datetime format
        into datetime objects. Otherwise, leave them encoded as numbers.
    autoclose : bool, optional
        If True, automatically close files to avoid OS Error of too many files
        being open.  However, this option doesn't work with streams, e.g.,
        BytesIO.
    concat_characters : bool, optional
        If True, concatenate along the last dimension of character arrays to
        form string arrays. Dimensions will only be concatenated over (and
        removed) if they have no corresponding variable and if they are only
        used as the last dimension of character arrays.
    decode_coords : bool, optional
        If True, decode the 'coordinates' attribute to identify coordinates in
        the resulting dataset.
    engine : {'netcdf4', 'scipy', 'pydap', 'h5netcdf', 'pynio', 'pseudonetcdf'}, optional
        Engine to use when reading files. If not provided, the default engine
        is chosen based on available dependencies, with a preference for
        'netcdf4'.
    chunks : int or dict, optional
        If chunks is provided, it used to load the new dataset into dask
        arrays. ``chunks={}`` loads the dataset with dask using a single
        chunk for all arrays.
    lock : False, True or threading.Lock, optional
        If chunks is provided, this argument is passed on to
        :py:func:`dask.array.from_array`. By default, a global lock is
        used when reading data from netCDF files with the netcdf4 and h5netcdf
        engines to avoid issues with concurrent access when using dask's
        multithreaded backend.
    cache : bool, optional
        If True, cache data loaded from the underlying datastore in memory as
        NumPy arrays when accessed to avoid reading from the underlying data-
        store multiple times. Defaults to True unless you specify the `chunks`
        argument to use dask, in which case it defaults to False. Does not
        change the behavior of coordinates corresponding to dimensions, which
        always load their data from disk into a ``pandas.Index``.
    drop_variables: string or iterable, optional
        A variable or list of variables to exclude from being parsed from the
        dataset. This may be useful to drop variables with problems or
        inconsistent values.
    backend_kwargs: dictionary, optional
        A dictionary of keyword arguments to pass on to the backend. This
        may be useful when backend options would improve performance or
        allow user control of dataset processing.
    sections : dict, optional
        Sections for calibration. The dictionary should contain key-var couples
        in which the key is the name of the calibration temp time series. And
        the var is a list of slice objects as 'slice(start, stop)'; start and
        stop in meter (float).
    Returns
    -------
    dataset : Dataset
        The newly created dataset.
    See Also
    --------
    open_mfdataset
    """

    xr_kws = inspect.signature(xr.open_dataset).parameters.keys()

    xr_kwargs = {k: v for k, v in kwargs.items() if k in xr_kws}
    ds_kwargs = {k: v for k, v in kwargs.items() if k not in xr_kws}

    ds_xr = xr.open_dataset(filename_or_obj, **xr_kwargs)

    ds = DataStore(data_vars=ds_xr.data_vars,
                   coords=ds_xr.coords,
                   attrs=ds_xr.attrs,
                   **ds_kwargs)
    return ds


def read_xml_dir(filepath,
                 timezone_netcdf='UTC',
                 timezone_ultima_xml='Europe/Amsterdam',
                 file_ext='*.xml',
                 **kwargs):

    filelist = sorted(glob.glob(os.path.join(filepath, file_ext)))
    array, timearr, meta, extra = grab_data(filelist)

    double_ended_flag = 'REV-ST' in array.dtype.names

    coords = {'x': array['LAF'][:, 0]}
    tcoords = coords_time(double_ended_flag, extra, timearr,
                          timezone_netcdf=timezone_netcdf,
                          timezone_ultima_xml=timezone_ultima_xml)
    coords.update(tcoords)

    dataset_dict = {}
    for name in array.dtype.names:
        if name in ['TMP', 'ST', 'AST', 'REV-ST', 'REV-AST']:
            dataset_dict[name] = (['x', 'time'], array[name])

        elif name == 'LAF':
            continue

        else:
            print(name)
            assert 0

    for key, item in extra.items():
        dataset_dict[key] = (['time'], item['array'])

    ds = DataStore(data_vars=dataset_dict,
                   coords=coords,
                   attrs=meta,
                   **kwargs)

    print(ds)

    return ds



# coding=utf-8
import hashlib
import os
import tempfile

import numpy as np

from dtscalibration import DataStore
from dtscalibration import open_datastore
from dtscalibration import read_xml_dir
from dtscalibration.datastore_utils import read_data_from_fp_numpy

fn = ["channel 1_20170921112245510.xml",
      "channel 1_20170921112746818.xml",
      "channel 1_20170921112746818.xml"]

if 1:
    # working dir is tests
    wd = os.path.dirname(os.path.abspath(__file__))
    data_dir = os.path.join(wd, 'data')

else:
    # working dir is src
    data_dir = os.path.join('..', '..', 'tests', 'data')


def test_read_data_from_single_file():
    """
    Check if read data from file is correct
    :return:
    """
    fp0 = os.path.join(data_dir, fn[0])
    data = read_data_from_fp_numpy(fp0)

    nx, ncols = data.shape

    err_msg = 'Not all points along the cable are read from file'
    np.testing.assert_equal(nx, 2330, err_msg=err_msg)

    err_msg = 'Not all columns are read from file'
    np.testing.assert_equal(ncols, 6, err_msg=err_msg)

    actual_hash = hashlib.sha1(data).hexdigest()
    desired_hash = '51b94dedd77c83c6cdd9dd132f379a39f742edae'

    assert actual_hash == desired_hash, 'The data is not read correctly'
    pass


def test_empty_construction():
    ds = DataStore()
    assert ds._initialized, 'Empty obj in not initialized'
    pass


def test_has_sectionattr_upon_creation():
    ds = DataStore()
    assert hasattr(ds, '_sections')
    assert isinstance(ds._sections, str)
    pass


def test_sections_property():
    ds = DataStore()
    sections1 = {
        'probe1Temperature': [slice(7.5, 17.), slice(70., 80.)],  # cold bath
        'probe2Temperature': [slice(24., 34.), slice(85., 95.)],  # warm bath
        }
    sections2 = {
        'probe1Temperature': [slice(0., 17.), slice(70., 80.)],  # cold bath
        'probe2Temperature': [slice(24., 34.), slice(85., 95.)],  # warm bath
        }
    ds.sections = sections1

    assert isinstance(ds._sections, str)

    assert ds.sections == sections1
    assert ds.sections != sections2

    # delete property
    del ds.sections
    assert ds.sections is None

    pass


def test_io_sections_property():
    ds = DataStore()
    sections = {
        'probe1Temperature': [slice(7.5, 17.), slice(70., 80.)],  # cold bath
        'probe2Temperature': [slice(24., 34.), slice(85., 95.)],  # warm bath
        }

    ds.sections = sections
    with tempfile.NamedTemporaryFile() as tmp:
        ds.to_netcdf(path=tmp.name)

        ds2 = open_datastore(tmp.name)

        assert ds.sections == ds2.sections

    pass


def test_read_xml_dir():
    filepath = data_dir
    ds = read_xml_dir(filepath,
                      timezone_netcdf='UTC',
                      timezone_ultima_xml='Europe/Amsterdam',
                      file_ext='*.xml')

    assert ds._initialized

    pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    return io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
        ).read()


setup(
    name='dtscalibration',
    version='0.1.2',
    license='BSD 3-Clause License',
    description='A Python package to load raw DTS files, perform a calibration, and plot the '
                'result',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
        ),
    author='Bas des Tombe',
    author_email='bdestombe@gmail.com',
    url='https://github.com/bdestombe/python-dts-calibration',
    packages=find_packages('src'),
    package_dir={
        '': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities',
        ],
    keywords=[
        'DTS', 'Calibration',
        ],
    install_requires=[
        'numpy',
        'xarray',
        'pyyaml',
        'xmltodict',
        'scipy'
        # eg: 'aspectlib==1.1.1', 'six>=1.7',
        ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
        },
    entry_points={
        'console_scripts': [
            'dtscalibration = dtscalibration.cli:main',
            ]
        },
    )
